@@ -11,6 +11,13 @@ from .decorators import login_required
 from . import emails
 from . import signals
 from .translations import gettext as _
+import sys
+
+# Handle Python 2.x and Python 3.x
+try:
+    from urllib.parse import quote      # Python 3.x
+except ImportError:
+    from urllib import quote            # Python 2.x
 
 def confirm_email(token):
     """ Verify email confirmation token and activate the user account."""
@@ -23,15 +30,15 @@ def confirm_email(token):
 
     if has_expired:
         flash(_('Your confirmation token has expired.'), 'error')
-        return redirect(user_manager.login_url)
+        return redirect(url_for('user.login'))
 
     if not is_valid:
         flash(_('Invalid confirmation token.'), 'error')
-        return redirect(user_manager.login_url)
+        return redirect(url_for('user.login'))
 
     # Confirm email by setting User.active=True and User.confirmed_at=utcnow()
     if db_adapter.UserEmailClass:
-        user_email = user_manager.find_user_email_by_id(object_id)
+        user_email = user_manager.get_user_email_by_id(object_id)
         if user_email:
             user_email.confirmed_at = datetime.utcnow()
             user = user_email.user
@@ -39,7 +46,7 @@ def confirm_email(token):
             user = None
     else:
         user_email = None
-        user = user_manager.find_user_by_id(object_id)
+        user = user_manager.get_user_by_id(object_id)
         user.confirmed_at = datetime.utcnow()
 
     if user:
@@ -47,19 +54,20 @@ def confirm_email(token):
         db_adapter.commit()
     else:                                               # pragma: no cover
         flash(_('Invalid confirmation token.'), 'error')
-        return redirect(user_manager.login_url)
+        return redirect(url_for('user.login'))
 
     # Send email_confirmed signal
     signals.user_confirmed_email.send(current_app._get_current_object(), user=user)
 
     # Prepare one-time system message
-    flash(_('Your email has been confirmed. Please sign in.'), 'success')
+    flash(_('Your email has been confirmed.'), 'success')
 
-    # Retrieve 'next' query parameter
-    next = request.args.get('next', '/')
-
-    # Redirect to the login page with the specified 'next' query parameter
-    return redirect(user_manager.login_url+'?next='+next)
+    # Redirect to login page or auto-login if configured
+    next = request.args.get('next', _endpoint_url(user_manager.after_confirm_endpoint))
+    if user_manager.auto_login_after_confirm:
+        return _do_login_user(user, next)  # Auto-login user after confirm
+    else:
+        return redirect(url_for('user.login')+'?next='+next)
 
 
 @login_required
@@ -70,7 +78,7 @@ def change_password():
 
     # Initialize form
     form = user_manager.change_password_form(request.form)
-    form.next.data = request.args.get('next', '/')  # Place ?next query param in next form field
+    form.next.data = request.args.get('next', _endpoint_url(user_manager.after_change_password_endpoint))  # Place ?next query param in next form field
 
     # Process valid POST
     if request.method=='POST' and form.validate():
@@ -105,7 +113,7 @@ def change_username():
 
     # Initialize form
     form = user_manager.change_username_form(request.form)
-    form.next.data = request.args.get('next', '/')  # Place ?next query param in next form field
+    form.next.data = request.args.get('next', _endpoint_url(user_manager.after_change_username_endpoint))  # Place ?next query param in next form field
 
     # Process valid POST
     if request.method=='POST' and form.validate():
@@ -139,10 +147,10 @@ def email_action(id, action):
     db_adapter = user_manager.db_adapter
 
     # Retrieve UserEmail by id
-    user_email = db_adapter.find_object(db_adapter.UserEmailClass, id=id)
+    user_email = db_adapter.find_first_object(db_adapter.UserEmailClass, id=id)
 
     # Users may only change their own UserEmails
-    if not user_email or user_email.user_id != current_user.id:
+    if not user_email or user_email.user_id != int(current_user.get_id()):
         return unauthorized()
 
     if action=='delete':
@@ -155,7 +163,7 @@ def email_action(id, action):
 
     elif action=='make-primary':
         # Disable previously primary emails
-        user_emails = db_adapter.find_object_list(db_adapter.UserEmailClass, user_id=current_user.id)
+        user_emails = db_adapter.find_all_objects(db_adapter.UserEmailClass, user_id=int(current_user.get_id()))
         for ue in user_emails:
             if ue.is_primary:
                 ue.is_primary = False
@@ -165,7 +173,7 @@ def email_action(id, action):
         db_adapter.commit()
 
     elif action=='confirm':
-        send_confirm_email_or_registered_email(user_email.user, user_email)
+        _send_confirm_email(user_email.user, user_email)
 
     else:
         return unauthorized()
@@ -188,7 +196,7 @@ def forgot_password():
         user, user_email = user_manager.find_user_by_email(email)
         if user:
             # Generate reset password link
-            token = user_manager.generate_token(user.id)
+            token = user_manager.generate_token(int(user.get_id()))
             reset_password_link = url_for('user.reset_password', token=token, _external=True)
 
             # Send forgot password email
@@ -206,7 +214,7 @@ def forgot_password():
         flash(_("A reset password email has been sent to '%(email)s'. Open that email and follow the instructions to reset your password.", email=email), 'success')
 
         # Redirect to the login page
-        return redirect(url_for('user.login'))
+        return redirect(_endpoint_url(user_manager.after_forgot_password_endpoint))
 
     # Process GET or invalid POST
     return render_template(user_manager.forgot_password_template, form=form)
@@ -217,17 +225,19 @@ def login():
     user_manager =  current_app.user_manager
     db_adapter = user_manager.db_adapter
 
-    next = request.args.get('next', '/')                  # Get the ?next=... query param
+    next = request.args.get('next', _endpoint_url(user_manager.after_login_endpoint))
+    reg_next = request.args.get('reg_next', _endpoint_url(user_manager.after_register_endpoint))
 
     # Immediately redirect already logged in users
-    if user_manager.auto_login:
-        if current_user.is_authenticated():
-            return redirect(next)
+    if current_user.is_authenticated() and user_manager.auto_login_at_login:
+        return redirect(next)
 
     # Initialize form
-    login_form = user_manager.login_form(request.form)
-    login_form.next.data = next                           # set hidden form field 'next'
-    register_form = user_manager.register_form()          # for login_or_register.html
+    login_form = user_manager.login_form(request.form)          # for login.html
+    register_form = user_manager.register_form()                # for login_or_register.html
+    if request.method!='POST':
+        login_form.next.data     = register_form.next.data = next
+        login_form.reg_next.data = register_form.reg_next.data = reg_next
 
     # Process valid POST
     if request.method=='POST' and login_form.validate():
@@ -238,8 +248,8 @@ def login():
             user = user_manager.find_user_by_username(login_form.username.data)
             user_email = None
             if user and db_adapter.UserEmailClass:
-                user_email = db_adapter.find_object(db_adapter.UserEmailClass,
-                        user_id=user.id,
+                user_email = db_adapter.find_first_object(db_adapter.UserEmailClass,
+                        user_id=int(user.get_id()),
                         is_primary=True,
                         )
             if not user and user_manager.enable_email:
@@ -250,23 +260,15 @@ def login():
 
         if user:
             if user.active:
-                # Use Flask-Login to sign in user
-                login_user(user)
-
-                # Send user_logged_in signal
-                signals.user_logged_in.send(current_app._get_current_object(), user=user)
-
-                # Prepare one-time system message
-                flash(_('You have signed in successfully.'), 'success')
-
-                # Redirect to 'next' URL
-                return redirect(login_form.next.data)
+                return _do_login_user(user, login_form.next.data, login_form.remember_me.data)  # Login user after Login
             else:
-                confirmed_at = user_email.confirmed_at if db_adapter.UserEmailClass else user.confirmed_at
+                confirmed_at = user_email.confirmed_at if user_email else user.confirmed_at
                 if user_manager.enable_confirm_email and not confirmed_at:
-                    flash(_('Your email address has not yet been confirmed. Check your email Inbox and Spam folders for the confirmation email or <a href="'+url_for('user.resend_confirm_email')+'">Re-send confirmation email</a>.'), 'error')
+                    url = url_for('user.resend_confirm_email')
+                    flash(_('Your email address has not yet been confirmed. Check your email Inbox and Spam folders for the confirmation email or <a href="%(url)s">Re-send confirmation email</a>.', url=url), 'error')
                 else:
                     flash(_('Your account has been disabled.'), 'error')
+                return redirect(url_for('user.home'))
 
     # Process GET or invalid POST
     return render_template(user_manager.login_template,
@@ -288,7 +290,7 @@ def logout():
     flash(_('You have signed out successfully.'), 'success')
 
     # Redirect to logout_next endpoint or '/'
-    next = request.args.get('next', '/')  # Get 'next' query param
+    next = request.args.get('next', _endpoint_url(user_manager.after_logout_endpoint))  # Get 'next' query param
     return redirect(next)
 
 
@@ -297,13 +299,13 @@ def manage_emails():
     user_manager =  current_app.user_manager
     db_adapter = user_manager.db_adapter
 
-    user_emails = db_adapter.find_object_list(db_adapter.UserEmailClass, user_id=current_user.id)
+    user_emails = db_adapter.find_all_objects(db_adapter.UserEmailClass, user_id=int(current_user.get_id()))
     form = user_manager.add_email_form()
 
     # Process valid POST request
     if request.method=="POST" and form.validate():
         user_emails = db_adapter.add_object(db_adapter.UserEmailClass,
-                user_id=current_user.id,
+                user_id=int(current_user.get_id()),
                 email=form.email.data)
         db_adapter.commit()
         return redirect(url_for('user.manage_emails'))
@@ -319,9 +321,15 @@ def register():
     user_manager =  current_app.user_manager
     db_adapter = user_manager.db_adapter
 
+    next = request.args.get('next', _endpoint_url(user_manager.after_login_endpoint))
+    reg_next = request.args.get('reg_next', _endpoint_url(user_manager.after_register_endpoint))
+
     # Initialize form
-    register_form = user_manager.register_form(request.form)
-    login_form = user_manager.login_form()          # for login_or_register.html
+    login_form = user_manager.login_form()                      # for login_or_register.html
+    register_form = user_manager.register_form(request.form)    # for register.html
+    if request.method!='POST':
+        login_form.next.data    = register_form.next.data = next
+        login_form.reg_next.data = register_form.reg_next.data = reg_next
 
     # Process valid POST
     if request.method=='POST' and register_form.validate():
@@ -380,13 +388,27 @@ def register():
 
         db_adapter.commit()
 
-        send_confirm_email_or_registered_email(user, user_email)
+        # Send 'registered' email and delete new User object if send fails
+        if user_manager.send_registered_email:
+            try:
+                # Send 'registered' email
+                _send_registered_email(user, user_email)
+            except Exception as e:
+                # delete new User object if send  fails
+                db_adapter.delete_object(user)
+                db_adapter.commit()
+                raise e
 
         # Send user_registered signal
         signals.user_registered.send(current_app._get_current_object(), user=user)
 
-        # Redirect to the login page
-        return redirect(url_for('user.login'))
+        # Redirect to login page or auto-login if configured
+        reg_next = register_form.reg_next.data
+        if user_manager.enable_confirm_email:
+            return redirect(reg_next)
+        if user_manager.auto_login_after_register:
+            return _do_login_user(user, reg_next)  # Auto-login user after Register
+        return redirect(_endpoint_url(user_manager.after_register_endpoint))
 
     # Process GET or invalid POST
     return render_template(user_manager.register_template,
@@ -394,6 +416,7 @@ def register():
             login_form=login_form,
             register_form=register_form)
 
+
 def resend_confirm_email():
     """Prompt for email and re-send email conformation email."""
     user_manager =  current_app.user_manager
@@ -409,10 +432,10 @@ def resend_confirm_email():
         # Find user by email
         user, user_email = user_manager.find_user_by_email(email)
         if user:
-            send_confirm_email_or_registered_email(user, user_email)
+            _send_confirm_email(user, user_email)
 
         # Redirect to the login page
-        return redirect(url_for('user.login'))
+        return redirect(_endpoint_url(user_manager.after_resend_confirm_email_endpoint))
 
     # Process GET or invalid POST
     return render_template(user_manager.resend_confirm_email_template, form=form)
@@ -430,13 +453,13 @@ def reset_password(token):
 
     if has_expired:
         flash(_('Your reset password token has expired.'), 'error')
-        return redirect(user_manager.login_url)
+        return redirect(url_for('user.login'))
 
     if not is_valid:
         flash(_('Your reset password token is invalid.'), 'error')
-        return redirect(user_manager.login_url)
+        return redirect(url_for('user.login'))
 
-    user = user_manager.find_user_by_id(user_id)
+    user = user_manager.get_user_by_id(user_id)
     if user:
         # Avoid re-using old tokens
         if hasattr(user, 'reset_password_token'):
@@ -445,7 +468,7 @@ def reset_password(token):
             verified = True
     if not user or not verified:
         flash(_('Your reset password token is invalid.'), 'error')
-        return redirect(user_manager.login_url)
+        return redirect(_endpoint_url(user_manager.login_endpoint))
 
     # Initialize form
     form = user_manager.reset_password_form(request.form)
@@ -478,60 +501,27 @@ def _get_email_address(user):
     user_manager =  current_app.user_manager
     db_adapter = user_manager.db_adapter
     if db_adapter.UserEmailClass:
-        user_email = db_adapter.find_object(db_adapter.UserEmailClass,
-                user_id=user.id,
+        user_email = db_adapter.find_first_object(db_adapter.UserEmailClass,
+                user_id=int(user.get_id()),
                 is_primary=True,
                 )
         return user_email.email if user_email else None
     else:
         return user.email
 
-def send_confirm_email_or_registered_email(user, user_email):
-    user_manager =  current_app.user_manager
-    db_adapter = user_manager.db_adapter
-
-    # Send 'confirm_email' or 'registered' email
-    if user_manager.enable_email:
-        email_address = user_email.email if user_email else user.email
-
-        try:
-            if user_manager.enable_confirm_email:
-                # Send 'confirm_email' email
-
-                # Generate confirm email link
-                object_id = user_email.id if user_email else user.id
-                token = user_manager.generate_token(object_id)
-                confirm_email_link = url_for('user.confirm_email', token=token, _external=True)
-
-                # Send email
-                emails.send_confirm_email_email(user, user_email, confirm_email_link)
-            else:
-                if user_manager.send_registered_email:
-                    # Send 'registered' email
-                    emails.send_registered_email(user, user_email)
-
-        except Exception as e:
-            # delete newly registered user if send email fails
-            db_adapter.delete_object(user)
-            db_adapter.commit()
-            raise e
-
-    # Prepare one-time system message
-    if user_manager.enable_confirm_email:
-        flash(_('A confirmation email has been sent to %(email)s with instructions to complete your registration.', email=email_address), 'success')
-    else:
-        flash(_('You have registered successfully. Please sign in.'), 'success')
-
 
 def unauthenticated():
     """ Prepare a Flash message and redirect to USER_UNAUTHENTICATED_URL"""
     # Prepare Flash message
-    url = request.script_root + request.path
+    url = request.url
     flash(_("You must be signed in to access '%(url)s'.", url=url), 'error')
 
+    # quote the fully qualified url
+    quoted_url = quote(url)
+
     # Redirect to USER_UNAUTHENTICATED_URL
     user_manager = current_app.user_manager
-    return redirect(user_manager.unauthenticated_url+'?next='+url)
+    return redirect(_endpoint_url(user_manager.unauthenticated_endpoint)+'?next='+ quoted_url)
 
 def unauthorized():
     """ Prepare a Flash message and redirect to USER_UNAUTHORIZED_URL"""
@@ -541,4 +531,68 @@ def unauthorized():
 
     # Redirect to USER_UNAUTHORIZED_URL
     user_manager = current_app.user_manager
-    return redirect(user_manager.unauthorized_url)
+    return redirect(_endpoint_url(user_manager.unauthorized_endpoint))
+
+def _send_registered_email(user, user_email):
+    user_manager =  current_app.user_manager
+    db_adapter = user_manager.db_adapter
+
+    # Send 'confirm_email' or 'registered' email
+    if user_manager.enable_email and user_manager.enable_confirm_email:
+        # Generate confirm email link
+        object_id = user_email.id if user_email else int(user.get_id())
+        token = user_manager.generate_token(object_id)
+        confirm_email_link = url_for('user.confirm_email', token=token, _external=True)
+
+        # Send email
+        emails.send_registered_email(user, user_email, confirm_email_link)
+
+        # Prepare one-time system message
+        if user_manager.enable_confirm_email:
+            email = user_email.email if user_email else user.email
+            flash(_('A confirmation email has been sent to %(email)s with instructions to complete your registration.', email=email), 'success')
+        else:
+            flash(_('You have registered successfully.'), 'success')
+
+
+def _send_confirm_email(user, user_email):
+    user_manager =  current_app.user_manager
+    db_adapter = user_manager.db_adapter
+
+    # Send 'confirm_email' or 'registered' email
+    if user_manager.enable_email and user_manager.enable_confirm_email:
+        # Generate confirm email link
+        object_id = user_email.id if user_email else int(user.get_id())
+        token = user_manager.generate_token(object_id)
+        confirm_email_link = url_for('user.confirm_email', token=token, _external=True)
+
+        # Send email
+        emails.send_confirm_email_email(user, user_email, confirm_email_link)
+
+        # Prepare one-time system message
+        email = user_email.email if user_email else user.email
+        flash(_('A confirmation email has been sent to %(email)s with instructions to complete your registration.', email=email), 'success')
+
+
+def _do_login_user(user, next, remember_me=False):
+    if user and user.is_active:
+        # Use Flask-Login to sign in user
+        #print('login_user: remember_me=', remember_me)
+        login_user(user, remember=remember_me)
+
+        # Send user_logged_in signal
+        signals.user_logged_in.send(current_app._get_current_object(), user=user)
+
+        # Prepare one-time system message
+        flash(_('You have signed in successfully.'), 'success')
+
+        # Redirect to 'next' URL
+        return redirect(next)
+
+    return unauthenticated()
+
+def _endpoint_url(endpoint):
+    url = '/'
+    if endpoint:
+        url = url_for(endpoint)
+    return url
\ No newline at end of file
